/*
 * Copyright 2012 Peter Lebbing <peter@digitalbrains.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "common.h"
#include "picprog.h"
#include "suart.h"
#include "avr_proto.h"

#ifndef TIFR0
#	define TIFR0 TIFR
#endif

static const char
greeting[] PROGMEM = AVR_GREETING;

static uint8_t clock_delay;
static uint8_t clock_scaler;

static uint8_t us_delay;
static uint8_t us_scaler;

/*
 * Selecting a clock outside the approx range 11 kHz to 26 MHz means we can't do
 * the timeout like we do here. Just reject it; choose a crystal in the range.
 */
#if ((F_CPU / 102400) == 0)
#	error "Clock freq too low for timeout!"
#elif ((F_CPU / 102400) > 255)
#	error "Clock freq too high for timeout!"
#endif
static const uint8_t
timeout_delay = F_CPU / 102400; /* Timeout 10 ms */

static const uint8_t
timeout_scaler = _BV(CS02) | _BV(CS00); /* Prescaler /1024 */

/* 10 MHz is maximum from datasheet, approach that figure */
static const uint32_t
default_clock_delay = F_CPU / 10000000UL;

static void
set_timer_rate(uint8_t *timer_val, uint8_t *scaler, uint32_t delay)
{
	uint8_t leadz, shift;

	if (delay == 0)
		delay = 1;

	leadz = __builtin_clzl(delay);
	if (leadz < 14) {
		return; /* Delay too large, bail */
	} else if (leadz < 16) {
		shift = 10;
		*scaler = _BV(CS02) | _BV(CS00);
	} else if (leadz < 18) {
		shift = 8;
		*scaler = _BV(CS02);
	} else if (leadz < 21) {
		shift = 6;
		*scaler = _BV(CS01) | _BV(CS00);
	} else if (leadz < 24) {
		shift = 3;
		*scaler = _BV(CS01);
	} else {
		shift = 0;
		*scaler = _BV(CS00);
	}
	*timer_val = delay >> shift;

	/*
	 * Round up in case of shifts, so it will be no shorter than given
	 * delay
	 */
	if (shift != 0)
		(*timer_val)++;
}

static uint32_t
get_timer_rate(const uint8_t timer_val, const uint8_t scaler)
{
	uint8_t shift;

	if (scaler == (_BV(CS02) | _BV(CS00)))
		shift = 10;
	else if (scaler == _BV(CS02))
		shift = 8;
	else if (scaler == (_BV(CS01) | _BV(CS00)))
		shift = 6;
	else if (scaler == _BV(CS01))
		shift = 3;
	else if (scaler == _BV(CS00))
		shift = 0;
	else
		return -1; /* Reserved value, return error */

	if (timer_val != 0)
		return (((uint32_t) timer_val) << shift);
	else
		return (0x100UL << shift);
}

static inline void
start_timer(uint8_t scaler)
{
	TCCR0B = 0;		/* Stop running */
	TCNT0 = 0;
	TIFR0 = _BV(OCF0A);
	TCCR0B = scaler;
}

static uint8_t
read_from_pc(uint8_t *buf, uint8_t count)
{
	uint8_t i;

	i = 0;
	OCR0A = timeout_delay;
	while (i < count) {
		start_timer(timeout_scaler);
		do {
		} while (suart_get_would_block() && bit_is_clear(TIFR0, OCF0A));

		if (bit_is_set(TIFR0, OCF0A))
			/* Timeout */
			break;

		buf[i++] = suart_get();
	}

	return i;
}

static void
do_special(uint8_t comm)
{
	PGM_P str;
	uint8_t c, ret, i;
	uint8_t buf[4];
	uint32_t *buf32;

	buf32 = (uint32_t *) buf;

	switch (comm) {

	case AVR_SPECIAL_GREET:
		suart_disable_rx();
		str = greeting;
		do {
			c = pgm_read_byte(str);
			str++;
			suart_put(c);
		} while (c != '\0');
		suart_enable_rx();
		break;

	case AVR_SPECIAL_SET_VPP_VDD:
		ret = read_from_pc(&c, 1);
		if (ret < 1)
			/* Timeout */
			return;
		if (c & AVR_VPP)
			PIC_VPP_PORT &= ~_BV(PIC_VPP);
		else
			PIC_VPP_PORT |= _BV(PIC_VPP);
		if (c & AVR_VDD)
			PIC_VDD_PORT |= _BV(PIC_VDD);
		else
			PIC_VDD_PORT &= ~_BV(PIC_VDD);
		suart_disable_rx();
		suart_put(0);
		suart_enable_rx();
		break;
	case AVR_SPECIAL_GET_CRYSTAL:
		*buf32 = F_CPU;
		suart_disable_rx();
		for (i = 0; i < 4; i++)
			suart_put(buf[i]);
		suart_enable_rx();
		break;
	case AVR_SPECIAL_SET_PIC_CLOCK:
		ret = read_from_pc(buf, 4);
		if (ret < 4)
			return; /* Timeout */
		if (*buf32 == 0xffffffff)
			*buf32 = default_clock_delay; /* Use default */
		set_timer_rate(&clock_delay, &clock_scaler, *buf32);
		*buf32 = get_timer_rate(clock_delay, clock_scaler);
		suart_disable_rx();
		for (i = 0; i < 4; i++)
			suart_put(buf[i]);
		suart_enable_rx();
		break;
	default:
		/* Just ignore it, they'll stop babbling soon enough :) */
		break;
	}
}

static void
send_pic_comm(uint8_t comm)
{
	uint8_t i;

	OCR0A = clock_delay;
	i = 6;
	PIC_DATA_DDR |= _BV(PIC_DATA);
	while (1) {
		PIC_CLOCK_PORT |= _BV(PIC_CLOCK);
		if (comm & 1)
			PIC_DATA_PORT |= _BV(PIC_DATA);
		else
			PIC_DATA_PORT &= ~_BV(PIC_DATA);
		start_timer(clock_scaler);
		comm >>= 1;
		loop_until_bit_is_set(TIFR0, OCF0A);
		PIC_CLOCK_PORT &= ~_BV(PIC_CLOCK);
		start_timer(clock_scaler);
		i--;
		if (i == 0)
			break;
		loop_until_bit_is_set(TIFR0, OCF0A);
	}
}

static void
send_pic_data(uint16_t data)
{
	uint8_t i;

	OCR0A = clock_delay;
	i = 16;
	PIC_DATA_DDR |= _BV(PIC_DATA);
	while (1) {
		PIC_CLOCK_PORT |= _BV(PIC_CLOCK);
		if (data & 1)
			PIC_DATA_PORT |= _BV(PIC_DATA);
		else
			PIC_DATA_PORT &= ~_BV(PIC_DATA);
		start_timer(clock_scaler);
		data >>= 1;
		loop_until_bit_is_set(TIFR0, OCF0A);
		PIC_CLOCK_PORT &= ~_BV(PIC_CLOCK);
		start_timer(clock_scaler);
		i--;
		if (i == 0)
			break;
		loop_until_bit_is_set(TIFR0, OCF0A);
	}
	loop_until_bit_is_set(TIFR0, OCF0A);
}

static uint16_t
read_pic_reply(void)
{
	uint8_t i;
	uint16_t data;

	OCR0A = clock_delay;
	i = 16;
	data = 0;
	PIC_DATA_DDR &= ~_BV(PIC_DATA);
	while (1) {
		PIC_CLOCK_PORT |= _BV(PIC_CLOCK);
		start_timer(clock_scaler);
		data >>= 1;
		loop_until_bit_is_set(TIFR0, OCF0A);
		PIC_CLOCK_PORT &= ~_BV(PIC_CLOCK);
		start_timer(clock_scaler);
		if (bit_is_set(PIC_DATA_PIN, PIC_DATA))
			data |= (1 << 15);
		i--;
		if (i == 0)
			break;
		loop_until_bit_is_set(TIFR0, OCF0A);
	}
	loop_until_bit_is_set(TIFR0, OCF0A);

	return data;
}

int
main(void)
{
	uint8_t i, c, kind, comm;
	uint16_t data;

#if USE_EEPROM_OSCCAL
	i = eeprom_read_byte(0);
	OSCCAL = i;
#endif

	/* Set trigger conditions for SUART RxD */
	EICRA = _BV(ISC01);

	set_timer_rate(&clock_delay, &clock_scaler, default_clock_delay);

	/* A microsecond delay between commands and data */
	set_timer_rate(&us_delay, &us_scaler, F_CPU / 1000000UL);

	PIC_VPP_DDR |= _BV(PIC_VPP);
	PIC_VDD_DDR |= _BV(PIC_VDD);
	PIC_CLOCK_DDR |= _BV(PIC_CLOCK);

	PIC_VPP_PORT |= _BV(PIC_VPP);

	suart_init(SUART_BAUD_TO_BRR(115200));
	sei();

	while (1) {
		c = suart_get();
		kind = c & AVR_KIND_MASK;
		comm = c & ~AVR_KIND_MASK;

		switch (kind) {
		case AVR_COMM_SPECIAL:
			do_special(comm);
			break;
		case AVR_COMM_SINGLE:
			suart_disable_rx();
			send_pic_comm(comm);
			OCR0A = us_delay;
			start_timer(us_scaler);
			suart_put(0);
			suart_enable_rx();
			loop_until_bit_is_set(TIFR0, OCF0A);
			break;
		case AVR_COMM_READ:
			suart_disable_rx();
			send_pic_comm(comm);
			OCR0A = us_delay;
			start_timer(us_scaler);
			loop_until_bit_is_set(TIFR0, OCF0A);
			data = read_pic_reply();
			OCR0A = us_delay;
			start_timer(us_scaler);
			suart_put(data & 0xff);
			suart_put(data >> 8);
			suart_enable_rx();
			loop_until_bit_is_set(TIFR0, OCF0A);
			break;
		case AVR_COMM_WRITE:
			i = read_from_pc((uint8_t *) & data, 2);
			if (i < 2)
				continue; /* Timeout */
			suart_disable_rx();
			send_pic_comm(comm);
			OCR0A = us_delay;
			start_timer(us_scaler);
			loop_until_bit_is_set(TIFR0, OCF0A);
			send_pic_data(data);
			OCR0A = us_delay;
			start_timer(us_scaler);
			suart_put(0);
			suart_enable_rx();
			loop_until_bit_is_set(TIFR0, OCF0A);
			break;
		}
	}
}
